/*
	Class_DataList.h
	Written by Hiep Dam
	From The Witches' Brew
	starlabs@aol.com
	Nov 20 1995
	
	Public domain.

	Notes:
	This is similar to a container class, except that the entries in
	the DataList are variable-sized entities. Note that in order to
	use variable-sized data, the DataList requires that each entry
	be prefixed by a length data type. This is most useful for
	string-like data types - although they *don't* have to be strings.
	
	A prime example of the variable-sized data that DataList uses are
	pascal strings, which are a stream of chars prefixed by a length
	byte. Because it's only a byte, the string is limited to 255
	chars. However, this need not be hard-coded. You can tell DataList
	what size the length prefix is, thereby setting your own string
	length limits. If for example, you use an unsigned long as a
	length prefix, you can use strings up to over 4 million chars in
	length! (Of course, the 4 million limit is purely in theory; in
	practice it's significantly smaller...)
	
	All of this makes DataList an extremely useful tool for dealing
	with variable-sized data...
	
	Note: This class has NOT been thoroughly checked for bugs. You
	have been forewarned...
	
	
	Format of a DataList:
	=====================
	n bytes		Number of entries
				[n is defined by size of NumEntriesType]
	----------------------------------------------------
	(variable)	Entry 1
	----------------------------------------------------
	(variable)	Entry 2
	----------------------------------------------------
	(variable)	Entry 3
	...
	...
	
	The format of an Entry is:
	==========================
	m byte(s)	Size (length) of first Entry
				[m is defined by size of EntryLengthType]
	----------------------------------------------------
	(variable)	Entry data

	
	
	For example, this could be substantiated to a STR# resource format,
	done by using DataList<short, char>...
	
	Result:
	
	2 bytes		Number of strings					<short>
	--------------------------------------------
	1 byte		Length of first string (1...255)	<char>
	(variable)	First string data					(variable)
	--------------------------------------------
	1 byte		Length of second string				<char>
	(variable)	Second string data					(variable)
	...
	...
	
	This, of course, would limit each string to 255 chars. If you
	wanted longer strings, try using DataList<short, unsigned long>.
	Now that's a lot of characters!

	2 bytes		Number of strings
	----------------------------------
	4 bytes		Length of first string (1..4,294,967,295!)
	(variable)	First string data
	----------------------------------
	4 bytes		Length of second string
	(variable)	Second string data
	...
	...
	
	
	
	*** Remember, indices into the entries start at 1, NOT 0! ***
	*** This makes it compatible with STR# resource usage.    ***
	
	
	Version History:
		1.0		12/2/95
			Hey, it works!
*/

typedef char *EntryPtr;
typedef char **EntryHdl;
typedef char *DataPtr;
typedef unsigned short EntryIndex;	// 1..number of entries

template<class NumEntriesType, class EntryLengthType>
class DataList {
	public:
		DataList();
		~DataList();
	
		virtual void		SetData(EntryHdl data);	// Makes a copy of the data
		virtual	EntryHdl	GetData();				// Returns the actual data (not a copy!)
		virtual NumEntriesType	GetNumberOfEntries();
		
		virtual void		SetCurrentEntryIndex(EntryIndex newIndex);
		virtual EntryPtr	GetCurrentEntry();
		virtual EntryPtr	GetNextEntry();	// Like GetCurrentEntry(), but advances index
		
		virtual EntryPtr	GetEntry(EntryIndex index);
		virtual void		SetEntry(EntryPtr entry, EntryIndex index);

		// These routines normally do not set the entry index, BUT if the
		// entry index is invalid (i.e. 0), the index is reset back to 1.
		virtual void		InsertEntry(EntryPtr entry, EntryIndex index);
		virtual void		AppendEntry(EntryPtr entry);
		
		// Always resets the current entry index to 1.
		virtual void		DeleteEntry(EntryIndex index);
		
		// Unlike the above routines, which expect the entry data to be
		// in Entry format (i.e. entry length, then entry data), these
		// routines will build the data in entry format for you.
		// Argument <length> is in bytes.
		virtual void		SetEntry(DataPtr entryData, EntryLengthType length, EntryIndex index);
		virtual void		InsertEntry(DataPtr entryData, EntryLengthType length, EntryIndex index);
		virtual void		AppendEntry(DataPtr entryData, EntryLengthType length);

	protected:
		EntryIndex	fCurrentEntryIndex;		// Current index, used by GetNextEntry() et al.
		EntryPtr	fCurrentEntryAddress;	// Address of current index entry (tricky tho)
		EntryHdl	fData;					// Handle to our datalist.
	
	private:
		virtual void		SetNumberOfEntries(EntryIndex numEntries);
};